(ql:quickload '(:cl-astar :trivial-benchmark))


(unless (alexandria:length= 2 (uiop:command-line-arguments))
  (format t "USAGE: sbcl --load benchmark.lisp <maze size> <repetitions>")
  (uiop:quit 1))

(declaim (type alexandria:array-length *size*))
(defparameter *size* (parse-integer (first (uiop:command-line-arguments))))

(declaim (type (simple-array character) *maze* *path*))
(defvar *maze*)

(defconstant maze-file "/tmp/maze.lisp")
;; (setf *random-state* (make-random-state t))
(if (uiop:file-exists-p maze-file)
    (with-open-file (maze maze-file)
      (setf *maze*
            (alexandria:copy-array (read maze)
                                   :element-type 'character
                                   :fill-pointer nil
                                   :adjustable nil)))
    (loop :initially (setf *maze* (make-array (list *size* *size*)
                                              :element-type 'character
                                              :initial-element #\Space))
          :for i :from 0 :below *size* :do
             (loop :for j :from 0 :below *size* :do
                      (when (zerop (random 5))
                        (setf (aref *maze* i j) #\*)))
          :finally (with-open-file (maze maze-file :direction :output)
                     (write *maze* :stream maze))))

(a*:define-path-finder find-path ()
    (:variables ((result (alexandria:copy-array
                          *maze*
                          :element-type 'character))
                 (size *size*)
                 (maze *maze*))
     :world-size (the alexandria:array-length (* size size))
     :coordinate-type a*:integer-coordinate
     :coordinate-encoder a*:encode-integer-coordinates
     :coordinate-decoder a*:decode-integer-coordinates
     :indexer (a*:make-row-major-indexer size)
     :goal-reached-p a*:goal-reached-exact-p
     :neighbour-enumerator (a*:make-4-directions-enumerator
                            :max-x size :max-y size)
     :exact-cost
     (lambda (x1 y1 x2 y2)
       (declare (ignorable x1 y1))
       (if (eql (aref maze y2 x2) #\Space)
           0.0
           most-positive-single-float))
     :heuristic-cost (a*:make-manhattan-distance-heuristic)
     :path-processor (lambda (x y) (setf (aref result y x) #\x))
     :path-finalizer (lambda () result))
  (declare (optimize (speed 3) (safety 0) (debug 0) (compilation-speed 0))
           (type alexandria:array-index size)
           (type (simple-array character) maze result)))

(compile 'find-path)

;; (disassemble 'find-path)

(let ((path (find-path 0 0 (1- *size*) (1- *size*)))) ;; warmup
  (declare (ignorable path))
  ;; (print path)
  )

(trivial-benchmark:with-timing
    ((parse-integer (second (uiop:command-line-arguments))))
  (find-path 0 0 (1- *size*) (1- *size*)))

