(in-package #:cl-astar/tests)


(define-test required-macro-arguments
  :parent cl-astar
  (fail
   (eval
    '(define-path-finder find-path ()
      (:indexer (make-row-major-indexer 10)
       :goal-reached-p goal-reached-exact-p
       :neighbour-enumerator (make-4-directions-enumerator :max-x 10 :max-y 10)
       :exact-cost (constantly 0.0)
       :heuristic-cost (make-manhattan-distance-heuristic)
       :path-processor (constantly nil)))))
  (fail
   (eval
    '(define-path-finder find-path ()
      (:world-size 100
       :goal-reached-p goal-reached-exact-p
       :neighbour-enumerator (make-4-directions-enumerator :max-x 10 :max-y 10)
       :exact-cost (constantly 0.0)
       :heuristic-cost (make-manhattan-distance-heuristic)
       :path-processor (constantly nil)))))
  (fail
   (eval
    '(define-path-finder find-path ()
      (:world-size 100
       :indexer (make-row-major-indexer 10)
       :neighbour-enumerator (make-4-directions-enumerator :max-x 10 :max-y 10)
       :exact-cost (constantly 0.0)
       :heuristic-cost (make-manhattan-distance-heuristic)
       :path-processor (constantly nil)))))
  (fail
   (eval
    '(define-path-finder find-path ()
      (:world-size 100
       :indexer (make-row-major-indexer 10)
       :goal-reached-p goal-reached-exact-p
       :exact-cost (constantly 0.0)
       :heuristic-cost (make-manhattan-distance-heuristic)
       :path-processor (constantly nil)))))
  (fail
   (eval
    '(define-path-finder find-path ()
      (:world-size 100
       :indexer (make-row-major-indexer 10)
       :goal-reached-p goal-reached-exact-p
       :neighbour-enumerator (make-4-directions-enumerator :max-x 10 :max-y 10)
       :heuristic-cost (make-manhattan-distance-heuristic)
       :path-processor (constantly nil)))))
  (fail
   (eval
    '(define-path-finder find-path ()
      (:world-size 100
       :indexer (make-row-major-indexer 10)
       :goal-reached-p goal-reached-exact-p
       :neighbour-enumerator (make-4-directions-enumerator :max-x 10 :max-y 10)
       :exact-cost (constantly 0.0)
       :path-processor (constantly nil)))))
  (fail
   (eval
    '(define-path-finder find-path ()
      (:world-size 100
       :indexer (make-row-major-indexer 10)
       :goal-reached-p goal-reached-exact-p
       :neighbour-enumerator (make-4-directions-enumerator :max-x 10 :max-y 10)
       :exact-cost (constantly 0.0)
       :heuristic-cost (make-manhattan-distance-heuristic)))))
  (finish
   (eval
    '(define-path-finder find-path ()
      (:world-size 100
       :indexer (make-row-major-indexer 10)
       :goal-reached-p goal-reached-exact-p
       :neighbour-enumerator (make-4-directions-enumerator :max-x 10 :max-y 10)
       :exact-cost (constantly 0.0)
       :heuristic-cost (make-manhattan-distance-heuristic)
       :path-processor (constantly nil))))))

(define-test excess-body
  :parent cl-astar
  (fail
   (eval
    '(define-path-finder find-path ()
      (:world-size 100
       :indexer (make-row-major-indexer 10)
       :goal-reached-p goal-reached-exact-p
       :neighbour-enumerator (make-4-directions-enumerator :max-x 10 :max-y 10)
       :exact-cost (constantly 0.0)
       :heuristic-cost (make-manhattan-distance-heuristic)
       :path-processor (constantly nil))
      (declare (optimize (speed 3)))
      "Function to find path"
      (print "it works!")))))

(define-test docstring
  :parent cl-astar
  (eval
   '(define-path-finder find-path ()
     (:world-size 100
      :indexer (make-row-major-indexer 10)
      :goal-reached-p goal-reached-exact-p
      :neighbour-enumerator (make-4-directions-enumerator :max-x 10 :max-y 10)
      :exact-cost (constantly 0.0)
      :heuristic-cost (make-manhattan-distance-heuristic)
      :path-processor (constantly nil))
     "find-path function"))
  (is string= "find-path function" (documentation #'find-path 'function)))

(define-test declaration
  :parent cl-astar
  (let (conditions)
    (handler-bind ((t (lambda (c) (push c conditions))))
      (eval
       `(define-path-finder ,(gensym "FIND-PATH-IGN") (ignored)
         (:world-size 100
          :coordinate-type integer-coordinate
          :coordinate-encoder encode-integer-coordinates
          :coordinate-decoder decode-integer-coordinates
          :indexer (make-row-major-indexer 10)
          :goal-reached-p goal-reached-exact-p
          :neighbour-enumerator (make-4-directions-enumerator
                                 :max-x 10 :max-y 10)
          :exact-cost (constantly 0.0)
          :heuristic-cost (make-manhattan-distance-heuristic)
          :path-processor (constantly nil))
         (declare (ignore ignored)))))
    (false conditions)))

(define-test too-many-docstrings
  :parent cl-astar
  (fail
   (eval
    '(define-path-finder find-path ()
      (:world-size 100
       :indexer (make-row-major-indexer 10)
       :goal-reached-p goal-reached-exact-p
       :neighbour-enumerator (make-4-directions-enumerator :max-x 10 :max-y 10)
       :exact-cost (constantly 0.0)
       :heuristic-cost (make-manhattan-distance-heuristic)
       :path-processor (constantly nil))
      "find-path function"
      "yeah"))))

(define-test parameters
  :parent cl-astar
  (finish
   (eval
    '(define-path-finder find-path (width height)
      (:world-size 100
       :indexer (make-row-major-indexer width)
       :goal-reached-p (lambda (x1 y1 x2 y2)
                         (and (= x1 width) (= y1 height)
                              (= x2 width) (= y2 height)))
       :neighbour-enumerator (make-4-directions-enumerator
                              :max-x width :max-y height)
       :exact-cost (constantly (/ 1.0 (* width height)))
       :heuristic-cost (make-manhattan-distance-heuristic (* width height))
       :path-initiator (constantly (* width height))
       :path-processor (constantly (* width height))
       :path-finalizer (constantly (* width height)))))))

(define-test variables
  :parent cl-astar
  (finish
   (eval
    '(define-path-finder find-path ()
      (:variables ((width 10) (height 10))
       :world-size 100
       :indexer (make-row-major-indexer width)
       :goal-reached-p (lambda (x1 y1 x2 y2)
                         (and (= x1 width) (= y1 height)
                              (= x2 width) (= y2 height)))
       :neighbour-enumerator (make-4-directions-enumerator
                              :max-x width :max-y height)
       :exact-cost (constantly (/ 1.0 (* width height)))
       :heuristic-cost (make-manhattan-distance-heuristic (* width height))
       :path-initiator (constantly (* width height))
       :path-processor (constantly (* width height))
       :path-finalizer (constantly (* width height)))))))

(define-test negative-cost
  :parent cl-astar
  (fail
   (eval
    '(define-path-finder find-path ()
      (:variables ((width 10) (height 10))
       :world-size (* width height)
       :coordinate-type integer-coordinate
       :coordinate-encoder encode-integer-coordinates
       :coordinate-decoder decode-integer-coordinates
       :indexer (make-row-major-indexer 10)
       :goal-reached-p goal-reached-exact-p
       :neighbour-enumerator (make-4-directions-enumerator :max-x 10 :max-y 10)
       :exact-cost (constantly -1.0)
       :heuristic-cost (make-manhattan-distance-heuristic)
       :path-processor (constantly nil))))
      warning))

(define-test smoke
  :parent cl-astar
  (eval
   '(define-path-finder find-path ()
     (:world-size 100
      :indexer (make-row-major-indexer 10)
      :goal-reached-p goal-reached-exact-p
      :neighbour-enumerator (make-4-directions-enumerator :max-x 10 :max-y 10)
      :exact-cost (constantly 0.0)
      :heuristic-cost (make-manhattan-distance-heuristic)
      :path-processor (constantly nil))))
  (true (fboundp 'find-path)))
