(in-package #:cl-astar/tests)


(define-test straightforward-4-directions
  :parent cl-astar
  (eval
   '(define-path-finder find-path (result)
     (:world-size 100
      :coordinate-type integer-coordinate
      :coordinate-encoder encode-integer-coordinates
      :coordinate-decoder decode-integer-coordinates
      :indexer (make-row-major-indexer 10)
      :goal-reached-p goal-reached-exact-p
      :neighbour-enumerator (make-4-directions-enumerator :max-x 10 :max-y 10)
      :exact-cost (constantly 0.0)
      :heuristic-cost (make-manhattan-distance-heuristic)
      :path-processor (lambda (x y) (push (list x y) result))
      :path-finalizer (lambda () result))))
  (is equalp
      (loop :for i :from 9 :downto 0 :collect (list 0 i))
      (find-path 0 0 0 9))
  (is equalp
      (loop :for i :from 9 :downto 0 :collect (list i 0))
      (find-path 0 0 9 0))
  (is equalp
      '((9 9) (9 8) (9 7) (9 6) (9 5) (9 4) (9 3) (9 2) (9 1) (8 1) (7 1) (6 1)
        (5 1) (4 1) (3 1) (2 1) (1 1) (0 1) (0 0))
      (find-path 0 0 9 9))
  (is equalp
      '((0 0))
      (find-path 0 0 0 0))
  (is equalp
      '((5 5))
      (find-path 5 5 5 5))
  (false (find-path 0 0 10 10)))

(define-test straightforward-8-directions
  :parent cl-astar
  (eval
   '(define-path-finder find-path (result)
     (:world-size 100
      :coordinate-type integer-coordinate
      :coordinate-encoder encode-integer-coordinates
      :coordinate-decoder decode-integer-coordinates
      :indexer (make-row-major-indexer 10)
      :goal-reached-p goal-reached-exact-p
      :neighbour-enumerator (make-8-directions-enumerator :max-x 10 :max-y 10)
      :exact-cost (constantly 0.0)
      :heuristic-cost (make-octile-distance-heuristic)
      :path-processor (lambda (x y) (push (list x y) result))
      :path-finalizer (lambda () result))))
  (is equalp
      (loop :for i :from 9 :downto 0 :collect (list 0 i))
      (find-path 0 0 0 9))
  (is equalp
      (loop :for i :from 9 :downto 0 :collect (list i 0))
      (find-path 0 0 9 0))
  (is equalp
      (loop :for i :from 9 :downto 0 :collect (list i i))
      (find-path 0 0 9 9))
  (is equalp
      '((0 0))
      (find-path 0 0 0 0))
  (is equalp
      '((5 5))
      (find-path 5 5 5 5))
  (false (find-path 0 0 10 10)))

(define-test obstacle-4-directions
  :parent cl-astar
  (let* ((size 10)
         (maze (make-array (list size size)
                           :element-type 'character
                           :initial-contents '("    *     "
                                               "    *     "
                                               "    *     "
                                               "    *     "
                                               "    *     "
                                               "    *     "
                                               "    *     "
                                               "    *     "
                                               "    *     "
                                               "          "))))
    (define-path-finder find-path (result)
        (:world-size (* size size)
         :coordinate-type integer-coordinate
         :coordinate-encoder encode-integer-coordinates
         :coordinate-decoder decode-integer-coordinates
         :indexer (make-row-major-indexer size)
         :goal-reached-p goal-reached-exact-p
         :neighbour-enumerator (make-4-directions-enumerator
                                :max-x size :max-y size)
         :exact-cost (lambda (x1 y1 x2 y2)
                       (declare (ignorable x1 y1))
                       (if (eql (aref maze y2 x2) #\Space)
                           0.0
                           most-positive-single-float))
         :heuristic-cost (make-manhattan-distance-heuristic)
         :path-processor (lambda (x y) (push (list x y) result))
         :path-finalizer (lambda () result)))
    (is equalp
        '((9 0) (9 1) (9 2) (9 3) (9 4) (9 5) (9 6) (9 7) (9 8) (8 8) (7 8)
          (6 8) (5 8) (5 9) (4 9) (3 9) (3 8) (3 7) (3 6) (3 5) (3 4) (3 3)
          (3 2) (3 1) (3 0) (2 0) (1 0) (0 0))
        (find-path 0 0 9 0))))

(define-test obstacle-8-directions
  :parent cl-astar
  (let* ((size 10)
         (maze (make-array (list size size)
                           :element-type 'character
                           :initial-contents '("    *     "
                                               "    *     "
                                               "    *     "
                                               "    *     "
                                               "    *     "
                                               "    *     "
                                               "    *     "
                                               "    *     "
                                               "    *     "
                                               "          "))))
    (define-path-finder find-path (result)
        (:world-size (* size size)
         :coordinate-type integer-coordinate
         :coordinate-encoder encode-integer-coordinates
         :coordinate-decoder decode-integer-coordinates
         :indexer (make-row-major-indexer size)
         :goal-reached-p goal-reached-exact-p
         :neighbour-enumerator (make-8-directions-enumerator
                                :max-x size :max-y size)
         :exact-cost (lambda (x1 y1 x2 y2)
                       (declare (ignorable x1 y1))
                       (if (eql (aref maze y2 x2) #\Space)
                           0.0
                           most-positive-single-float))
         :heuristic-cost (make-octile-distance-heuristic)
         :path-processor (lambda (x y) (push (list x y) result))
         :path-finalizer (lambda () result)))
    (is equalp
        '((9 0) (9 1) (8 2) (7 3) (6 4) (6 5) (5 6) (5 7) (5 8) (4 9)
          (3 8) (3 7) (3 6) (3 5) (3 4) (3 3) (2 2) (1 1) (0 0))
        (find-path 0 0 9 0))))

(define-test concave-obstacle-4-directions
  :parent cl-astar
  (let* ((size 10)
         (maze (make-array (list size size)
                           :element-type 'character
                           :initial-contents '("          "
                                               " ******** "
                                               "        * "
                                               "        * "
                                               "        * "
                                               "        * "
                                               "        * "
                                               "        * "
                                               " ******** "
                                               "          "))))
    (define-path-finder find-path (result)
        (:world-size (* size size)
         :coordinate-type integer-coordinate
         :coordinate-encoder encode-integer-coordinates
         :coordinate-decoder decode-integer-coordinates
         :indexer (make-row-major-indexer size)
         :goal-reached-p goal-reached-exact-p
         :neighbour-enumerator (make-4-directions-enumerator
                                :max-x size :max-y size)
         :exact-cost (lambda (x1 y1 x2 y2)
                       (declare (ignorable x1 y1))
                       (if (eql (aref maze y2 x2) #\Space)
                           0.0
                           most-positive-single-float))
         :heuristic-cost (make-manhattan-distance-heuristic)
         :path-processor (lambda (x y) (push (list x y) result))
         :path-finalizer (lambda () result)))
    (is equalp
        '((9 4) (9 3) (9 2) (9 1) (9 0) (8 0) (7 0) (6 0) (5 0) (4 0) (3 0)
          (2 0) (1 0) (0 0) (0 1) (0 2) (0 3) (0 4))
        (find-path 0 4 9 4))))

(define-test concave-obstacle-8-directions
  :parent cl-astar
  (let* ((size 10)
         (maze (make-array (list size size)
                           :element-type 'character
                           :initial-contents '("          "
                                               " ******** "
                                               "        * "
                                               "        * "
                                               "        * "
                                               "        * "
                                               "        * "
                                               "        * "
                                               " ******** "
                                               "          "))))
    (define-path-finder find-path (result)
        (:world-size (* size size)
         :coordinate-type integer-coordinate
         :coordinate-encoder encode-integer-coordinates
         :coordinate-decoder decode-integer-coordinates
         :indexer (make-row-major-indexer size)
         :goal-reached-p goal-reached-exact-p
         :neighbour-enumerator (make-8-directions-enumerator
                                :max-x size :max-y size)
         :exact-cost (lambda (x1 y1 x2 y2)
                       (declare (ignorable x1 y1))
                       (if (eql (aref maze y2 x2) #\Space)
                           0.0
                           most-positive-single-float))
         :heuristic-cost (make-octile-distance-heuristic)
         :path-processor (lambda (x y) (push (list x y) result))
         :path-finalizer (lambda () result)))
    (is equalp
        '((9 4) (9 3) (9 2) (9 1) (8 0) (7 0) (6 0) (5 0) (4 0) (3 0) (2 0)
          (1 0) (0 1) (0 2) (0 3) (0 4))
        (find-path 0 4 9 4))))

(define-test no-path-4-directions
  :parent cl-astar
  (let* ((size 10)
         (maze (make-array (list size size)
                           :element-type 'character
                           :initial-contents '("    *     "
                                               "    *     "
                                               "    *     "
                                               "    *     "
                                               "    *     "
                                               "    *     "
                                               "    *     "
                                               "    *     "
                                               "    *     "
                                               "    *     "))))
    (define-path-finder find-path (result)
        (:world-size (* size size)
         :coordinate-type integer-coordinate
         :coordinate-encoder encode-integer-coordinates
         :coordinate-decoder decode-integer-coordinates
         :indexer (make-row-major-indexer size)
         :goal-reached-p goal-reached-exact-p
         :neighbour-enumerator (make-4-directions-enumerator
                                :max-x size :max-y size)
         :exact-cost (lambda (x1 y1 x2 y2)
                       (declare (ignorable x1 y1))
                       (if (eql (aref maze y2 x2) #\Space)
                           0.0
                           most-positive-single-float))
         :heuristic-cost (make-manhattan-distance-heuristic)
         :path-processor (lambda (x y) (push (list x y) result))
         :path-finalizer (lambda () result)))
    (false (find-path 0 0 9 0))))

(define-test no-path-8-directions
  :parent cl-astar
  (let* ((size 10)
         (maze (make-array (list size size)
                           :element-type 'character
                           :initial-contents '("    *     "
                                               "    *     "
                                               "    *     "
                                               "    *     "
                                               "    *     "
                                               "    *     "
                                               "    *     "
                                               "    *     "
                                               "    *     "
                                               "    *     "))))
    (define-path-finder find-path (result)
        (:world-size (* size size)
         :coordinate-type integer-coordinate
         :coordinate-encoder encode-integer-coordinates
         :coordinate-decoder decode-integer-coordinates
         :indexer (make-row-major-indexer size)
         :goal-reached-p goal-reached-exact-p
         :neighbour-enumerator (make-8-directions-enumerator
                                :max-x size :max-y size)
         :exact-cost (lambda (x1 y1 x2 y2)
                       (declare (ignorable x1 y1))
                       (if (eql (aref maze y2 x2) #\Space)
                           0.0
                           most-positive-single-float))
         :heuristic-cost (make-octile-distance-heuristic)
         :path-processor (lambda (x y) (push (list x y) result))
         :path-finalizer (lambda () result)))
    (false (find-path 0 0 9 0))))

(define-test snake-4-directions
  :parent cl-astar
  (let* ((size 10)
         (maze (make-array (list size size)
                           :element-type 'character
                           :initial-contents '("  *   *   "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "*   *   * "))))
    (define-path-finder find-path (result)
        (:world-size (* size size)
         :coordinate-type integer-coordinate
         :coordinate-encoder encode-integer-coordinates
         :coordinate-decoder decode-integer-coordinates
         :indexer (make-row-major-indexer size)
         :goal-reached-p goal-reached-exact-p
         :neighbour-enumerator (make-4-directions-enumerator
                                :max-x size :max-y size)
         :exact-cost (lambda (x1 y1 x2 y2)
                       (declare (ignorable x1 y1))
                       (if (eql (aref maze y2 x2) #\Space)
                           0.0
                           most-positive-single-float))
         :heuristic-cost (make-manhattan-distance-heuristic)
         :path-processor (lambda (x y) (push (list x y) result))
         :path-finalizer (lambda () result)))
    (is equalp
        '((9 9) (9 8) (9 7) (9 6) (9 5) (9 4) (9 3) (9 2) (9 1) (9 0) (8 0)
          (7 0) (7 1) (7 2) (7 3) (7 4) (7 5) (7 6) (7 7) (7 8) (7 9) (6 9)
          (5 9) (5 8) (5 7) (5 6) (5 5) (5 4) (5 3) (5 2) (5 1) (5 0) (4 0)
          (3 0) (3 1) (3 2) (3 3) (3 4) (3 5) (3 6) (3 7) (3 8) (3 9) (2 9)
          (1 9) (1 8) (1 7) (1 6) (1 5) (1 4) (1 3) (1 2) (1 1) (1 0) (0 0))
        (find-path 0 0 9 9))))

(define-test snake-8-directions
  :parent cl-astar
  (let* ((size 10)
         (maze (make-array (list size size)
                           :element-type 'character
                           :initial-contents '("  *   *   "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "*   *   * "))))
    (define-path-finder find-path (result)
        (:world-size (* size size)
         :coordinate-type integer-coordinate
         :coordinate-encoder encode-integer-coordinates
         :coordinate-decoder decode-integer-coordinates
         :indexer (make-row-major-indexer size)
         :goal-reached-p goal-reached-exact-p
         :neighbour-enumerator (make-8-directions-enumerator
                                :max-x size :max-y size)
         :exact-cost (lambda (x1 y1 x2 y2)
                       (declare (ignorable x1 y1))
                       (if (eql (aref maze y2 x2) #\Space)
                           0.0
                           most-positive-single-float))
         :heuristic-cost (make-octile-distance-heuristic)
         :path-processor (lambda (x y) (push (list x y) result))
         :path-finalizer (lambda () result)))
    (is equalp
        '((9 9) (9 8) (9 7) (9 6) (9 5) (9 4) (9 3) (9 2) (9 1) (8 0)
          (7 1) (7 2) (7 3) (7 4) (7 5) (7 6) (7 7) (7 8) (6 9)
          (5 8) (5 7) (5 6) (5 5) (5 4) (5 3) (5 2) (5 1) (4 0)
          (3 1) (3 2) (3 3) (3 4) (3 5) (3 6) (3 7) (3 8) (2 9)
          (1 8) (1 7) (1 6) (1 5) (1 4) (1 3) (1 2) (1 1) (0 0))
        (find-path 0 0 9 9))))

(define-test snake-euclidean
  :parent cl-astar
  (let* ((size 10)
         (maze (make-array (list size size)
                           :element-type 'character
                           :initial-contents '("  *   *   "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "*   *   * "))))
    (define-path-finder find-path (result)
        (:world-size (* size size)
         :coordinate-type integer-coordinate
         :coordinate-encoder encode-integer-coordinates
         :coordinate-decoder decode-integer-coordinates
         :indexer (make-row-major-indexer size)
         :goal-reached-p goal-reached-exact-p
         :neighbour-enumerator (make-8-directions-enumerator
                                :max-x size :max-y size)
         :exact-cost (lambda (x1 y1 x2 y2)
                       (declare (ignorable x1 y1))
                       (if (eql (aref maze y2 x2) #\Space)
                           0.0
                           most-positive-single-float))
         :heuristic-cost (make-euclidean-distance-heuristic)
         :path-processor (lambda (x y) (push (list x y) result))
         :path-finalizer (lambda () result)))
    (is equalp
        '((9 9) (9 8) (9 7) (9 6) (9 5) (9 4) (9 3) (9 2) (9 1) (8 0)
          (7 1) (7 2) (7 3) (7 4) (7 5) (7 6) (7 7) (7 8) (6 9)
          (5 8) (5 7) (5 6) (5 5) (5 4) (5 3) (5 2) (5 1) (4 0)
          (3 1) (3 2) (3 3) (3 4) (3 5) (3 6) (3 7) (3 8) (2 9)
          (1 8) (1 7) (1 6) (1 5) (1 4) (1 3) (1 2) (1 1) (0 0))
        (find-path 0 0 9 9))))

(define-test snake-column-major
  :parent cl-astar
  (let* ((size 10)
         (maze (make-array (list size size)
                           :element-type 'character
                           :initial-contents '("  *   *   "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "*   *   * "))))
    (define-path-finder find-path (result)
        (:world-size (* size size)
         :coordinate-type integer-coordinate
         :coordinate-encoder encode-integer-coordinates
         :coordinate-decoder decode-integer-coordinates
         :indexer (make-column-major-indexer size)
         :goal-reached-p goal-reached-exact-p
         :neighbour-enumerator (make-8-directions-enumerator
                                :max-x size :max-y size)
         :exact-cost (lambda (x1 y1 x2 y2)
                       (declare (ignorable x1 y1))
                       (if (eql (aref maze y2 x2) #\Space)
                           0.0
                           most-positive-single-float))
         :heuristic-cost (make-octile-distance-heuristic)
         :path-processor (lambda (x y) (push (list x y) result))
         :path-finalizer (lambda () result)))
    (is equalp
        '((9 9) (9 8) (9 7) (9 6) (9 5) (9 4) (9 3) (9 2) (9 1) (8 0)
          (7 1) (7 2) (7 3) (7 4) (7 5) (7 6) (7 7) (7 8) (6 9)
          (5 8) (5 7) (5 6) (5 5) (5 4) (5 3) (5 2) (5 1) (4 0)
          (3 1) (3 2) (3 3) (3 4) (3 5) (3 6) (3 7) (3 8) (2 9)
          (1 8) (1 7) (1 6) (1 5) (1 4) (1 3) (1 2) (1 1) (0 0))
        (find-path 0 0 9 9))))

(define-test snake-floating
  :parent cl-astar
  (let* ((size 10)
         (maze (make-array (list size size)
                           :element-type 'character
                           :initial-contents '("  *   *   "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "*   *   * "))))
    (define-path-finder find-path (result)
        (:world-size (* size size)
         :indexer (make-row-major-indexer size)
         :goal-reached-p goal-reached-exact-p
         :neighbour-enumerator (make-8-directions-enumerator
                                :max-x size :max-y size)
         :exact-cost (lambda (x1 y1 x2 y2)
                       (declare (ignorable x1 y1))
                       (if (eql (aref maze (floor y2) (floor x2)) #\Space)
                           0.0
                           most-positive-single-float))
         :heuristic-cost (make-octile-distance-heuristic)
         :path-processor (lambda (x y) (push (list x y) result))
         :path-finalizer (lambda () result)))
    (is equalp
        '((9 9) (9 8) (9 7) (9 6) (9 5) (9 4) (9 3) (9 2) (9 1) (8 0)
          (7 1) (7 2) (7 3) (7 4) (7 5) (7 6) (7 7) (7 8) (6 9)
          (5 8) (5 7) (5 6) (5 5) (5 4) (5 3) (5 2) (5 1) (4 0)
          (3 1) (3 2) (3 3) (3 4) (3 5) (3 6) (3 7) (3 8) (2 9)
          (1 8) (1 7) (1 6) (1 5) (1 4) (1 3) (1 2) (1 1) (0 0))
        (find-path 0.0 0.0 9.0 9.0))))

(define-test big-frontier
  :parent cl-astar
  (let* ((size 10)
         (maze (make-array (list size size)
                           :element-type 'character
                           :initial-contents '("  *   *   "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "* * * * * "
                                               "*   *   * "))))
    (define-path-finder find-path (result)
        (:frontier-size 2
         :world-size (* size size)
         :coordinate-type integer-coordinate
         :coordinate-encoder encode-integer-coordinates
         :coordinate-decoder decode-integer-coordinates
         :indexer (make-row-major-indexer size)
         :goal-reached-p goal-reached-exact-p
         :neighbour-enumerator (make-8-directions-enumerator
                                :max-x size :max-y size)
         :exact-cost (lambda (x1 y1 x2 y2)
                       (declare (ignorable x1 y1))
                       (if (eql (aref maze y2 x2) #\Space)
                           0.0
                           most-positive-single-float))
         :heuristic-cost (make-octile-distance-heuristic)
         :path-processor (lambda (x y) (push (list x y) result))
         :path-finalizer (lambda () result)))
    (is equalp
        '((9 9) (9 8) (9 7) (9 6) (9 5) (9 4) (9 3) (9 2) (9 1) (8 0)
          (7 1) (7 2) (7 3) (7 4) (7 5) (7 6) (7 7) (7 8) (6 9)
          (5 8) (5 7) (5 6) (5 5) (5 4) (5 3) (5 2) (5 1) (4 0)
          (3 1) (3 2) (3 3) (3 4) (3 5) (3 6) (3 7) (3 8) (2 9)
          (1 8) (1 7) (1 6) (1 5) (1 4) (1 3) (1 2) (1 1) (0 0))
        (find-path 0 0 9 9))))
